import io
import logging
from typing import Dict, List, Union

import pandas as pd
import pyarrow as pa
from pyarrow.lib import ArrowInvalid

from odp.tabular_v2.client import Cursor, Table
from odp.tabular_v2.client.validation import validate_data_against_schema


class Transaction:
    """
    a transaction is created implicitly when a table is used as a context manager:

        with table as tx:
            ...

    transaction should be used to modify the data, and make the modifications atomic (which means users won't see
    the changes while they are being made, but only all at once when the transaction is committed at the end).

    they transaction won't commit (and instead rollback) if an exception is raised inside the block.

    when a transaction is created, it might buffer some data locally to improve the performance of the system.
    """

    _buffer: pa.Table

    def __init__(self, table: Table, tx_id: str):
        if not tx_id:
            raise ValueError("tx_id must not be empty")
        self._table = table
        self._id = tx_id
        self._buffer = pa.Table.from_pylist([], schema=table.schema())

    def select(self, filter: str = "", vars: Union[Dict, List, None] = None) -> Cursor:
        """
        perform a select on the current transaction, which will include any changes made in the transaction so far.
        """
        self.flush()
        return self._table._query_cursor(type="select", filter=filter, vars=vars, tx_id=self._id)

    def replace(self, filter: str = "", vars: Union[Dict, List, None] = None, **kwargs) -> Cursor:
        """remove the rows that match the filter from the table and return them in a cursor.

        The rows can be just discarded, modified, re-inserted or anything else.
        """
        filter = filter or kwargs.get("query", "")
        if not filter:
            raise ValueError("For your own safety, a filter is required, use 1==1 to match all rows")
        self.flush()  # we flush any pending changes to be consistent later
        return self._table._query_cursor(type="replace", filter=filter, vars=vars, tx_id=self._id)

    def delete(self, query: str = "") -> int:
        """
        delete rows that match the query

        Note: similarly to the replace, some rows might be false positive and should be added back, but this
        happens internally and is not exposed to the user.
        Returns how many rows were changed
        """
        self.flush()  # we flush any pending changes to be consistent later
        ct = 0
        for b in self.replace(query).batches():  # NOTE(oha): we must consume the iterator to make it work
            ct += b.num_rows
        return ct

    def flush(self):
        """
        flush the data to the server, in case some data is buffered locally
        """
        logging.debug("flushing to stage %s", self._id)

        if self._buffer.num_rows == 0:
            return

        schema = self._buffer.schema
        buf = io.BytesIO()
        w = pa.ipc.RecordBatchStreamWriter(buf, schema, options=pa.ipc.IpcWriteOptions(compression="lz4"))

        # recursively split the batch in smaller ones if it's too big
        def write_batch(b: pa.RecordBatch):
            # small enough to be sent in one go
            if b.nbytes < 30_000_000 and b.num_rows < 5_000:
                w.write_batch(b)
            elif b.num_rows > 1:
                mid = b.num_rows // 2
                write_batch(b.slice(0, mid))
                write_batch(b.slice(mid))
            else:
                # we can't split it further
                w.write_batch(b)

        for b in self._buffer.to_batches():
            try:
                write_batch(b)
            except ArrowInvalid as e:
                raise ValueError("Invalid arrow format") from e
        w.close()

        self._table._impl(
            path="/api/table/v2/sdk/insert",
            params={
                "table_id": self._table._id,
                "tx_id": self._id,
            },
            data=buf.getvalue(),
        ).json()
        self._buffer = pa.Table.from_pylist([], schema=schema)

    def insert(self, data: Union[Dict, List[Dict], pa.RecordBatch, pd.DataFrame, pa.Table]):
        """
        add data to the internal buffer to be inserted into the table
        if the buffered data is enough, it will be automatically flushed

        accept a single dictionary, a list of dictionaries, a pandas DataFrame, or a pyarrow RecordBatch
        """

        schema = self._buffer.schema

        if isinstance(data, dict):
            validate_data_against_schema([data], schema)
            data = pa.Table.from_pylist([data], schema=schema)
        elif isinstance(data, list):
            validate_data_against_schema(data, schema)
            data = pa.Table.from_pylist(data, schema=schema)
        elif isinstance(data, pd.DataFrame):
            validate_data_against_schema(data, schema)
            data = pa.Table.from_pandas(data, schema=schema)
        elif isinstance(data, pa.RecordBatch):
            data = pa.Table.from_batches([data], schema=schema)
        elif isinstance(data, pa.Table):
            pass
        else:
            raise ValueError(f"unexpected type {type(data)}")

        self._buffer = pa.concat_tables([self._buffer, data]).combine_chunks()

        if self._buffer.num_rows >= 10_000 or self._buffer.nbytes >= 10_000_000:
            self.flush()



from typing import Dict

from ..resource import ResourceDto, ResourceSpec
from ..resource_registry import DEFAULT_RESOURCE_TYPE_REGISTRY
from ..validators import validate_resource_name


class ObservableSpec(ResourceSpec):
    observable_class: str
    """Qualified name of the observable class"""

    ref: str
    """Qualified name of associated dataset or data collection"""

    details: Dict
    """Full observable object"""

    _validate_ref = validate_resource_name(
        "ref",
    )
    _validate_observable_class = validate_resource_name(
        "observable_class",
    )


class ObservableDto(ResourceDto):
    _kind: str = "catalog.hubocean.io/observable"
    _version: str = "v1alpha2"

    spec: ObservableSpec


DEFAULT_RESOURCE_TYPE_REGISTRY.add(ObservableDto)



import io
import json
from typing import TYPE_CHECKING, Iterator, Optional, Union

import pyarrow as pa

if TYPE_CHECKING:
    from odp.tabular_v2.client import Table


class Raw:
    def __init__(self, table: "Table"):
        self.table = table

    def list(self, query: Optional[str] = None, vars: Optional[dict] = None) -> list:
        res = self.table._impl(
            "/api/table/v2/raw/list",
            params={"table_id": self.table._id, "query": query, "vars": vars},
            data={},
        )
        body = res.json()
        return body["files"]

    def list_batches(self, query: Optional[str] = None, vars: Optional[dict] = None) -> pa.Table:
        res = self.table._impl(
            "/api/table/v2/raw/batch-list",
            params={"table_id": self.table._id},
            data={"query": query, "vars": vars},
        )
        reader = pa.ipc.open_stream(res.reader())
        batches = [batch for batch in reader]
        return pa.Table.from_batches(batches, schema=reader.schema)

    def upload(self, name: str, data: Union[bytes, io.IOBase]) -> str:
        res = self.table._impl(
            "/api/table/v2/raw/upload",
            params={"table_id": self.table._id, "name": name},
            data=data,
        )
        body = res.json()
        return body["raw_id"]

    def update_meta(self, identifier: str, data: dict) -> dict:
        res = self.table._impl(
            "/api/table/v2/raw/update_meta",
            params={"table_id": self.table._id, "id": identifier},
            data=json.dumps(data).encode("utf-8"),
        )
        return res.json()

    def download(self, id: str) -> Iterator[bytes]:
        res = self.table._impl(
            "/api/table/v2/raw/download",
            params={"table_id": self.table._id, "id": id},
        )
        return res.iter()

    def delete(self, id: str) -> None:
        self.table._impl(
            "/api/table/v2/raw/delete",
            params={"table_id": self.table._id, "id": id},
        )

    def ingest(self, id: str) -> None:
        self.table._impl(
            "/api/table/v2/raw/ingest",
            params={"table_id": self.table._id, "id": id},
            data={},
        )



import io
import json
import logging
import os
from typing import TYPE_CHECKING, Dict, Iterator, Union

import requests
from requests.adapters import HTTPAdapter

# NOTE(oha): deferred imports to avoid circular dependencies, will be fixed when we remove the old client
# from odp.client.auth import InteractiveTokenProvider
# from odp.client.catalog_client import CatalogClient
from odp.tabular_v2.util import Iter2Reader

if TYPE_CHECKING:
    from dataset import Dataset


class _Response:
    # Abstraction for the response object, shared between http client and test client
    def __init__(self, res: Union[requests.Response, Iterator[bytes], Dict, bytes]):
        logging.info("response: %s", res)
        if isinstance(res, requests.Response):
            if res.status_code == 204:
                raise FileNotFoundError(res.text)
            res.raise_for_status()
        # logging.info("response: %s", res)
        self.res = res

    def reader(self):
        if isinstance(self.res, bytes):
            return io.BytesIO(self.res)
        if isinstance(self.res, Iterator):
            return Iter2Reader(self.res)
        return self.res.raw

    def iter(self) -> Iterator[bytes]:
        if isinstance(self.res, bytes):
            return iter([self.res])
        if isinstance(self.res, Iterator):
            return self.res
        return self.res.iter_content()

    def all(self) -> bytes:
        if isinstance(self.res, bytes):
            return self.res
        if isinstance(self.res, Iterator):
            return b"".join(self.res)
        return self.res.content

    def json(self) -> dict:
        if self.res is None:
            return None
        if isinstance(self.res, dict):
            return self.res
        return self.res.json()


class Client:
    """
    New ODP client for accessing datasets and other resources.
    supersedes the old odp.OdpClient
    """

    def __init__(self, base_url: str = "https://api.hubocean.earth/", jwt_bearer: str = "", api_key: str = ""):
        if api_key:
            self._auth = lambda: f"ApiKey {api_key}"
        elif jwt_bearer:
            if not jwt_bearer.startswith("Bearer "):
                jwt_bearer = "Bearer " + jwt_bearer
            self._auth = lambda: jwt_bearer
        elif os.getenv("JUPYTERHUB_API_TOKEN"):

            def get_token():
                res = requests.post("http://localhost:8000/access_token")
                res.raise_for_status()
                token: str = res.json()["token"]
                return "Bearer " + token

            self._auth = get_token
        else:
            self._setup_jwt()

        self.base_url = base_url.rstrip("/")

        # FIXME: this is wrong, sessions can't be this long
        self._http_client = requests.Session()
        self._http_client.headers.setdefault("User-Agent", "odp-sdk-python")
        from odp.client.catalog_client import CatalogClient

        self._catalog = CatalogClient(http_client=self._http_client)

    def _setup_jwt(self):
        """
        fire up the JWT setup process on the browser
        """
        client_id = os.getenv("ODP_CLIENT_ID", "f96fc4a5-195b-43cc-adb2-10506a82bb82")
        from odp.client.auth import InteractiveTokenProvider

        prov = InteractiveTokenProvider(client_id=client_id)
        self._auth = lambda: f"Bearer {prov.get_token()}"
        # raise NotImplementedError("JWT authentication is not implemented yet.")

    def _request(self, req: requests.Request) -> requests.Response:
        """
        base implementation to send a request to ODP
        may be overridden to use mechanisms other than http
        """
        req.headers.setdefault("Authorization", self._auth())  # allow override
        req.headers.setdefault("User-Agent", "odp-sdk-python")
        if isinstance(req.data, dict):
            req.data = json.dumps(req.data)
            req.headers.setdefault("Content-Type", "application/json")
        preq = req.prepare()
        # logging.info("request: %s %s %s", preq.method, preq.url, preq.body)

        with requests.Session() as s:
            adapter = HTTPAdapter(pool_connections=4, pool_maxsize=4, max_retries=8)
            s.mount("http://", adapter)
            s.mount("https://", adapter)
            res = s.send(preq, stream=True)
        logging.info("response: %s in %.2fs from %s", res.status_code, res.elapsed.total_seconds(), res.url)
        return res

    def dataset(self, id: str) -> "Dataset":
        """fetch a dataset by id"""
        from .dataset import Dataset

        # NOTE(oha): ktable does not strictly need a catalog, and catalog v2 is not ready yet
        # we just silently pass a dataset with not catalog features for now, and add more later
        return Dataset(self, id)
        # req = requests.Request("GET", self.base_url + "/api/catalog/v2/datasets/" + safe_id(id))
        # req.timeout = 10
        # res = self._request(req)
        # if res.status_code in (204, 404):
        #    raise FileNotFoundError(f"Dataset {id} not found: %s %s", res.status_code, res.content)
        # return Dataset._from_catalog_v2(self, res.json())



from typing import List, Literal, Optional
from uuid import UUID

from ...util.cheapdantic import BaseModel
from .tabular_store import TablePartitioningSpec


class TableSpec(BaseModel):
    table_schema: dict
    partitioning: Optional[List[TablePartitioningSpec]] = None


class StageDataPoints(BaseModel):
    """Model for update data point endpoint."""

    action: Literal["create", "commit"]
    stage_id: Optional[UUID]

    # @root_validator(pre=True) # FIXME
    def _validate_action(cls, values):
        if values.get("action") == "create" and values.get("stage_id"):
            raise ValueError("stage id cannot be issued with create action")
        elif values.get("action") == "commit" and not values.get("stage_id"):
            raise ValueError("stage id must be issued with commit action")

        return values



import json
import logging
import re
from time import sleep
from typing import Dict, Iterable, List, Optional, Tuple, Union
from uuid import UUID
from warnings import warn

import requests

from ..util.cheapdantic import BaseModel
from .dto.table_spec import StageDataPoints, TableSpec
from .dto.tabular_store import TableStage
from .exc import OdpResourceExistsError, OdpResourceNotFoundError
from .http_client import OdpHttpClient
from .utils import convert_geometry
from .utils.ndjson import parse_ndjson
from odp.dto import DatasetDto

try:
    from pandas import DataFrame
except ImportError:
    DataFrame = ImportError
    warn("Pandas not installed. DataFrame support will not be available.")


class OdpTabularStorageClient(BaseModel):
    http_client: OdpHttpClient
    """HTTP-request client"""

    tabular_storage_endpoint: str = "/data"
    """Endpoint appended to base URL in order to interact with the ODP data-APIs"""

    pagination_size: int = 10_000
    """List-limit when paginating"""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        if not re.match(r"^/\w+(?<!/)", self.tabular_storage_endpoint):
            raise ValueError(f"Invalid endpoint: {self.tabular_storage_endpoint}")

    def tabular_endpoint(self, dataset: Union[str, DatasetDto], *path: str) -> str:
        """Get actual tabular endpoint given a dataset

        Args:
            dataset: Dataset to be used
            path: Path-components to be added to the endpoint

        Returns:
            Tabular endpoint given `dataset`
        """
        dataset_id = dataset.get_ref() if isinstance(dataset, DatasetDto) else dataset
        ret = f"{self.http_client.base_url}{self.tabular_storage_endpoint}/{dataset_id}"
        return "/".join([ret, *path])

    def create_schema(self, dataset: Union[str, DatasetDto], table_spec: TableSpec) -> TableSpec:
        """Create Schema

        Args:
            dataset: Dataset manifest or dataset ID
            table_spec: Specifications of the schema that is being created

        Returns:
            Specifications of the schema that is being created

        Raises:
            OdpResourceExistsError: If the schema already exists with the same identifier
        """

        response = self.http_client.post(self.tabular_endpoint(dataset, "schema"), content=table_spec)

        try:
            response.raise_for_status()
        except requests.HTTPError as e:
            if response.status_code == 409:
                raise OdpResourceExistsError("Schema with identifier already exists") from e
            raise

        return TableSpec(**response.json())

    def get_schema(self, dataset: Union[str, DatasetDto]) -> TableSpec:
        """Get schema

        Args:
            dataset: Dataset manifest or dataset ID

        Returns:
            Specifications of the schema that is being queried

        Raises
            OdpResourceNotFoundError: If the schema cannot be found
        """
        response = self.http_client.get(self.tabular_endpoint(dataset, "schema"))

        try:
            response.raise_for_status()
        except requests.HTTPError as e:
            if response.status_code == 404:
                raise OdpResourceNotFoundError("Schema not found") from e
            raise

        return TableSpec(**json.loads(response.text))

    def delete_schema(self, dataset: Union[str, DatasetDto], delete_data: bool = False):
        """Delete schema

        Args:
            dataset: Dataset manifest or dataset ID
            delete_data: Bool to specify whether the data should be deleted as well

        Raises:
            OdpResourceNotFoundError: If the schema cannot be found
        """

        response = self.http_client.delete(
            self.tabular_endpoint(dataset, "schema"), params={"delete_data": delete_data}
        )

        try:
            response.raise_for_status()
        except requests.HTTPError as e:
            if response.status_code == 404:
                raise OdpResourceNotFoundError("Schema not found") from e
            raise

    def create_stage_request(self, resource_dto: DatasetDto) -> TableStage:
        """Create Stage

        Args:
            resource_dto: Dataset manifest

        Returns:
            Specifications of the stage that is being created

        Raises
            Stage with the specified identifier already exists
        """
        stage_data = StageDataPoints(action="create", stage_id=None)
        response = self.http_client.post(self.tabular_endpoint(resource_dto, "stage"), content=stage_data)

        try:
            response.raise_for_status()
        except requests.HTTPError as e:
            if response.status_code == 409:
                raise OdpResourceExistsError("Stage with identifier already exists") from e
            raise

        return TableStage(**response.json())

    def commit_stage_request(self, resource_dto: DatasetDto, table_stage: TableStage):
        """Commit Stage

        Args:
            resource_dto: Dataset manifest
            table_stage: Stage specifications for the stage that is to be committed
        """
        stage_data = StageDataPoints(action="commit", stage_id=table_stage.stage_id)

        response = self.http_client.post(self.tabular_endpoint(resource_dto, "stage"), content=stage_data)
        response.raise_for_status()

    def get_stage_request(self, resource_dto: DatasetDto, stage: Union[UUID, TableStage]) -> TableStage:
        """Get Stage

        Args:
            resource_dto: Dataset manifest
            stage: Identifier for the stage

        Returns:
            Stage that is queried for

        Raises
            OdpResourceNotFoundError: If the schema cannot be found
        """
        stage = stage.stage_id if isinstance(stage, TableStage) else stage
        response = self.http_client.get(self.tabular_endpoint(resource_dto, "stage", str(stage)))

        try:
            response.raise_for_status()
        except requests.HTTPError as e:
            if response.status_code == 400:  # This is coming as 400 for both 400 and 404 from ODP
                raise OdpResourceNotFoundError("Schema not found") from e
            raise

        return TableStage(**response.json())

    def list_stage_request(self, resource_dto: DatasetDto) -> List[TableStage]:
        """List Stages for a dataset

        Args:
            resource_dto: Dataset manifest

        Returns:
            Stages that are related to the dataset

        Raises
            OdpResourceNotFoundError: If the schema cannot be found
        """
        response = self.http_client.get(self.tabular_endpoint(resource_dto, "stage"))

        try:
            response.raise_for_status()
        except requests.HTTPError as e:
            if response.status_code == 400:  # This is coming as 400 for both 400 and 404 from ODP
                raise OdpResourceNotFoundError("Schema not found") from e
            raise

        return [TableStage(**table_stage) for table_stage in response.json()]

    def delete_stage_request(self, resource_dto: DatasetDto, table_stage: TableStage, force_delete=False):
        """Delete Stage

        Args:
            resource_dto: Dataset manifest
            table_stage: Table stage that wants to be deleted
            force_delete: Bool to specify whether the data should be force deleted

        Raises
            OdpResourceNotFoundError: If the schema cannot be found
        """

        response = self.http_client.delete(
            self.tabular_endpoint(resource_dto, "stage", str(table_stage.stage_id)),
            params={"force_delete": force_delete},
        )

        try:
            response.raise_for_status()
        except requests.HTTPError as e:
            if response.status_code == 400:  # This is coming as 400 for both 400 and 404 from ODP
                raise OdpResourceNotFoundError("Schema not found") from e
            raise

    def select(
        self,
        resource_dto: DatasetDto,
        filter_query: Optional[dict] = None,
        limit: Optional[int] = None,
    ) -> Iterable[dict]:
        """Read data from tabular API

        Args:
            resource_dto: Dataset manifest
            filter_query: Filter query in OQS format
            limit: Limit for the number of rows returned

        Yields:
            Each row of the data
        """
        if limit and limit < 0:
            raise ValueError("Limit should be a positive")

        try:
            dataset_schema = self.get_schema(resource_dto)
        except OdpResourceNotFoundError:
            print(f"Schema not found for resource {resource_dto.metadata.name}: geometry conversion skipped")
            dataset_schema = None

        cursor = None
        while True:
            rows = self._select_page(dataset_schema, resource_dto, filter_query, limit, cursor)
            for row, is_meta in rows:
                if is_meta:
                    cursor = row.get("@@next")
                    continue

                yield row
                if limit:
                    limit -= 1
                    if limit <= 0:
                        return

            if not cursor:
                break

    def select_as_stream(
        self,
        resource_dto: DatasetDto,
        filter_query: Optional[dict] = None,
        limit: Optional[int] = None,
    ) -> Iterable[dict]:
        """Select data from dataset

        Args:
            resource_dto: Dataset manifest
            filter_query: Filter query in OQS format
            limit: limit for the number of rows returned

        Returns:
            Data that is queried as a stream

        Raises
            OdpResourceNotFoundError: If the schema cannot be found
        """
        yield from self.select(resource_dto, filter_query, limit=limit)

    def select_as_list(
        self,
        resource_dto: DatasetDto,
        filter_query: Optional[dict] = None,
        limit: Optional[int] = None,
    ) -> list[dict]:
        """Select data from dataset

        Args:
            resource_dto: Dataset manifest
            filter_query: Filter query in OQS format
            limit: limit for the number of rows returned

        Returns:
            Data that is queried as a list

        Raises
            OdpResourceNotFoundError: If the schema cannot be found
        """

        return list(self.select(resource_dto, filter_query, limit))

    def _select_page(
        self,
        dataset_schema: TableSpec,
        resource_dto: DatasetDto,
        filter_query: Optional[dict] = None,
        limit: Optional[int] = None,
        cursor: Optional[str] = None,
        result_geometry: Optional[str] = "geojson",
    ) -> Iterable[Tuple[dict, bool]]:
        """Method to query a specific page from the data"""
        query_parameters = {}
        if limit:
            query_parameters["limit"] = limit
        if cursor:
            query_parameters["cursor"] = cursor

        for retry_delay in [0.5, 2, 5, 20]:  # exponential backoff
            response = self.http_client.post(
                self.tabular_endpoint(resource_dto, "list"),
                params=query_parameters,
                content=filter_query,
                headers={"Accept": "application/x-ndjson"},
                stream=True,
            )
            try:
                response.raise_for_status()
            except requests.HTTPError as e:
                if response.status_code == 404:
                    raise OdpResourceNotFoundError("Resource not found") from e
                logging.warning("retrying in %s after fetch failed: %s", retry_delay, e)
                sleep(retry_delay)
                continue
            break

        for row in parse_ndjson(response.iter_content(chunk_size=None)):
            if len(row) == 1 and next(iter(row.keys())).startswith("@@"):
                is_meta = True
            else:
                is_meta = False
            if dataset_schema:
                row = self._convert_geometry_with_schema(dataset_schema, row, result_geometry)
            yield row, is_meta

    @staticmethod
    def _convert_geometry_with_schema(dataset_schema: TableSpec, row: dict, result_geometry: str) -> dict:
        geometry_cols = [
            column
            for column, column_data in dataset_schema.table_schema.items()
            if column_data.get("type") == "geometry"
        ]
        for col in geometry_cols:
            try:
                row[col] = convert_geometry(row[col], result_geometry)
            except KeyError:
                continue
        return row

    def select_as_dataframe(self, resource_dto: DatasetDto, filter_query: Optional[dict] = None) -> DataFrame:
        """
        Select data from dataset as a DataFrame

        Args:
            resource_dto: Dataset manifest
            filter_query: Filter query in OQS format

        Returns:
            Data that is queried in DataFrame format

        Raises
            OdpResourceNotFoundError: If the schema cannot be found
        """

        data = self.select_as_list(resource_dto, filter_query)

        return DataFrame(data)

    def write(self, resource_dto: DatasetDto, data: List[Dict], table_stage: Optional[TableStage] = None):
        """
        Write data to dataset

        Args:
            resource_dto: Dataset manifest
            data: Data to ingest
            table_stage: Stage specifications for the stage to ingest

        Raises
            OdpResourceNotFoundError: If the schema cannot be found
        """

        self._write_limited_size(self.tabular_endpoint(resource_dto), data, table_stage)

    def _write_limited_size(self, url: str, data: List[Dict], table_stage: Optional[TableStage] = None):
        if len(data) < 1:
            return

        body = dict()
        if table_stage:
            body["stage_id"] = str(table_stage.stage_id)
        body["data"] = data

        response = self.http_client.post(url, content=body)

        try:
            response.raise_for_status()
        except requests.HTTPError as e:
            if response.status_code == 404:
                raise OdpResourceNotFoundError("Resource not found") from e
            raise requests.HTTPError(f"HTTP Error - {response.status_code}: {response.text}")

    def write_dataframe(self, resource_dto: DatasetDto, data: DataFrame, table_stage: Optional[TableStage] = None):
        """
        Write data to dataset in DataFrame format

        Args:
            resource_dto: Dataset manifest
            data: Data to ingest in DataFrame format
            table_stage: Stage specifications for the stage to ingest

        Raises
            OdpResourceNotFoundError: If the schema cannot be found
        """

        data_list = data.to_dict(orient="records")

        self.write(resource_dto, data_list, table_stage)

    def delete(self, resource_dto: DatasetDto, filter_query: Optional[dict] = None):
        """
        Delete data from dataset

        Args:
            resource_dto: Dataset manifest
            filter_query: Filter query in OQS format

        Raises
            OdpResourceNotFoundError: If the schema cannot be found
        """
        response = self.http_client.post(self.tabular_endpoint(resource_dto, "delete"), content=filter_query)

        try:
            response.raise_for_status()
        except requests.HTTPError as e:
            if response.status_code == 404:
                raise OdpResourceNotFoundError("Resource not found") from e
            raise

    def update(self, resource_dto: DatasetDto, data: List[Dict], filter_query: dict):
        """
        Update data from dataset

        Args:
            resource_dto: Dataset manifest
            filter_query: Filter query in OQS format
            data: data to update

        Raises
            OdpResourceNotFoundError: If the schema cannot be found
        """
        response = self.http_client.patch(
            self.tabular_endpoint(resource_dto), content={"update_filters": filter_query, "data": data}
        )

        try:
            response.raise_for_status()
        except requests.HTTPError as e:
            if response.status_code == 404:
                raise OdpResourceNotFoundError("Resource not found") from e
            raise

    def update_dataframe(self, resource_dto: DatasetDto, data: DataFrame, filter_query: Optional[dict] = None):
        """
        Update data from dataset in DataFrame format

        Args:
            resource_dto: Dataset manifest
            filter_query: Filter query in OQS format
            data: Data to update in DataFrame format

        Raises
            OdpResourceNotFoundError: If the schema cannot be found
        """

        data_list = data.values.tolist()

        self.update(resource_dto, data_list, filter_query)



import logging
from typing import Optional

import requests

from odp.new_client import Client


# TEMPORARY
class Dataset:
    def __init__(self, cli: Client, data: dict):
        self.cli = cli
        self.uuid = data["id"]
        self.name = data.get("name", None)

    @classmethod
    def by_name(cls, cli: Client, name: str) -> Optional["Dataset"]:
        """
        Fetch a dataset by its name.
        Returns None if the dataset does not exist.
        """
        res = cli._request(
            requests.Request(
                method="GET",
                url=cli.base_url + "/api/catalog/v2/datasets",
            )
        )
        res.raise_for_status()
        for item in res.json():
            if item["name"] == name:
                return cls(cli, item)
        return None

    @classmethod
    def by_name_old(cls, cli: Client, name: str) -> Optional["Dataset"]:
        res = cli._request(
            requests.Request(
                method="GET",
                url=cli.base_url + "/catalog/catalog.hubocean.io/dataset/" + name,
            )
        )
        if res.status_code == 404:
            return None
        res.raise_for_status()
        return Dataset(cli, res.json())

    @classmethod
    def create(cls, cli: Client, name: str, description: str = "n/a") -> "Dataset":
        """
        Create a new dataset with the given name and description.
        Returns the created Dataset object.
        """
        res = cli._request(
            requests.Request(
                method="POST",
                url=cli.base_url + "/api/catalog/v2/datasets",
                data={
                    "name": name,
                    "description": description,
                },
            )
        )
        res.raise_for_status()
        return cls(cli, res.json())

    @classmethod
    def find_old(cls, cli: Client, name: Optional[str] = None) -> list["Dataset"]:
        conds = [{"#EQUALS": ["$kind", "catalog.hubocean.io/dataset"]}]
        if name is not None:
            conds.append({"#EQUALS": ["$metadata.name", name]})
        res = cli._request(
            requests.Request(
                method="POST",
                url=cli.base_url + "/catalog/list/",
                data={"#AND": conds},
            )
        )
        return [Dataset(cli, item) for item in res.json()["results"]]

    @classmethod
    def create_old(cls, cli: Client, name: str, description: str = "n/a") -> "Dataset":
        res = cli._request(
            requests.Request(
                method="POST",
                url=cli.base_url + "/catalog",
                data={
                    "kind": "catalog.hubocean.io/dataset",
                    "version": "v1alpha3",
                    "metadata": {
                        "name": name,
                        "description": description,
                        "labels": {},
                    },
                    "spec": {
                        "storage_controller": "registry.hubocean.io/storageController/storage-tabular",
                        "storage_class": "registry.hubocean.io/storageClass/tabular",
                        "maintainer": {
                            "contact": "Just Me <raw_client_example@hubocean.earth>"
                        },  # <-- strict syntax here
                    },
                },
            )
        )
        logging.info("created: %s", res.content)
        res.raise_for_status()
        return Dataset(cli, res.json())
